from __future__ import annotations

from typing import Awaitable

import pluggy

from moriarty.matrix.job_manager.params import InferenceJob, InferenceResult

project_name = "moriarty.matrix.bridge"
"""
The entry-point name of this extension.

Should be used in ``pyproject.toml`` as ``[project.entry-points."moriarty.matrix.bridge"]``
"""
hookimpl = pluggy.HookimplMarker(project_name)
"""
Hookimpl marker for this extension, extension module should use this marker

Example:

    .. code-block:: python

        @hookimpl
        def register(manager):
            ...
"""

hookspec = pluggy.HookspecMarker(project_name)


@hookspec
def register(manager):
    """
    For more information about this function, please check the :ref:`manager`

    We provided an example package for you in ``{project_root}/example/extension/custom-bridge``.

    Example:

    .. code-block:: python

        class CustomBridge(QueueBridge):
            register_name = "example"

        from moriarty.matrix.job_manager.bridge.plugin import hookimpl

        @hookimpl
        def register(manager):
            manager.register(CustomBridge)


    Config ``project.entry-points`` so that we can find it

    .. code-block:: toml

        [project.entry-points."moriarty.matrix.bridge"]
        {whatever-name} = "{package}.{path}.{to}.{file-with-hookimpl-function}"


    You can verify it by `moriarty-matrix list-plugins`.
    """


class QueueBridge:
    """
    A bridge is used to adapt different queue system to job manager

    Implementation of subclass should follow:
        - When dequeue a job or result, `process_func` will be called.
          And if exception occurs, the job may not be removed from queue.
          So that we can retry later.
    """

    register_name: str

    def __init__(
        self,
        bridge_result_queue_url: str = None,
        *args,
        **kwargs,
    ):
        self.bridge_result_queue_url = bridge_result_queue_url

    def make_job_queue_url(self, endpoint_name: str, priority: int = None) -> str:
        raise NotImplementedError

    def remove_job_queue(self, endpoint_name: str, priority: int = None) -> None:
        raise NotImplementedError

    async def list_avaliable_priorities(self, endpoint_name: str) -> list[int]:
        raise NotImplementedError

    async def enqueue_job(self, endpoint_name: str, job: InferenceJob, priority: int = None) -> str:
        raise NotImplementedError

    async def dequeue_job(
        self,
        endpoint_name: str,
        process_func: Awaitable[InferenceJob],
        size: int = 1,
        priority: int = None,
    ) -> int:
        raise NotImplementedError

    async def enqueue_result(self, result: InferenceResult) -> str:
        raise NotImplementedError

    async def dequeue_result(
        self, process_func: Awaitable[InferenceResult], size: int = 10
    ) -> list[InferenceResult]:
        raise NotImplementedError



from __future__ import annotations

import glob
import importlib
import os
from os.path import basename, dirname, isfile, join
from typing import Any

import pluggy
from fastapi import Depends

from moriarty.envs import get_spawner_name
from moriarty.exceptions import PluginInitializationError, PluginNotFoundError
from moriarty.log import logger
from moriarty.matrix.operator_.spawner import impl, plugin


def get_spawner_manager() -> SpawnerManager:
    return SpawnerManager()


async def get_spawner(
    spawner_name: str = Depends(get_spawner_name),
    spawner_manager: SpawnerManager = Depends(get_spawner_manager),
) -> plugin.Spawner:
    spawner = spawner_manager.init(spawner_name)
    await spawner.prepare()
    return spawner


class SpawnerManager:
    def __init__(self):
        self.pm = pluggy.PluginManager(plugin.project_name)
        self.pm.add_hookspecs(plugin)
        self._registed_cls: dict[str, type[plugin.Spawner]] = {}

        self.pm.load_setuptools_entrypoints(plugin.project_name)
        self.load_all_local_model()

    def load_all_local_model(self):
        """
        Import all implementation in ./impl
        """
        logger.debug(f"Loading all local model from {impl}")

        return self._load_dir(impl)

    @property
    def registed_cls(self) -> dict[str, type]:
        """
        Access all registed class.

        Lazy load, only load once.
        """
        if self._registed_cls:
            return self._registed_cls
        for f in self.pm.hook.register(manager=self):
            try:
                f()
            except Exception as e:
                logger.error(f"One spawner registration failed: {e}")
                logger.exception(e)
                continue
        return self._registed_cls

    def _load_dir(self, module):
        """
        Import all python files in a submodule.
        """
        modules = glob.glob(join(dirname(module.__file__), "*.py"))
        sub_packages = (
            basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")
        )
        packages = (str(module.__package__) + "." + i for i in sub_packages)
        for p in packages:
            logger.debug(f"Loading {p}")
            self.pm.register(importlib.import_module(p))

    def _normalize_name(self, name: str) -> str:
        return name.strip().lower()

    def register(self, cls: type[plugin.Spawner]):
        """
        Register a new module, if the module is already registed, skip it.
        """

        cls_name = self._normalize_name(cls.register_name)
        logger.debug(f"Register for new module: {cls_name}")
        if cls in self._registed_cls.values():
            logger.warning(
                f"SKIP: trying to register {cls} as '{cls_name}', but {cls_name} is already registed."
            )
            return
        if not issubclass(cls, plugin.Spawner):
            logger.error(f"SKIP: {cls_name} is not a subclass of {plugin.Spawner}")
            return
        self._registed_cls[cls_name] = cls

    def init(self, c, *args, **kwargs: dict[str, Any]) -> plugin.Spawner:
        """
        Init a new subclass of plugin.Spawner.

        Raises:
            NotFoundError: if cls_name is not registered
            InitializationError: if failed to initialize
        """
        if isinstance(c, plugin.Spawner):
            return c

        if isinstance(c, type):
            cls_type = c
        else:
            c = self._normalize_name(c)

            if not c in self.registed_cls:
                raise PluginNotFoundError(f"{c} is not registed.")
            cls_type = self.registed_cls[c]
        try:
            return cls_type(*args, **kwargs)
        except Exception as e:
            raise PluginInitializationError(e)



import pytest
from click.testing import CliRunner
from sqlalchemy import delete

from moriarty.matrix.job_manager.bridge_wrapper import BridgeWrapper
from moriarty.matrix.job_manager.params import InferenceJob
from moriarty.matrix.operator_.cli import drop
from moriarty.matrix.operator_.operator_ import Bridger
from moriarty.matrix.operator_.orm import EndpointORM


@pytest.fixture
async def mock_endpoint(async_session):
    name = "mock"
    async_session.add(
        EndpointORM(
            endpoint_name=name,
        )
    )
    await async_session.commit()
    yield name

    async_session.execute(delete(EndpointORM).where(EndpointORM.endpoint_name == name))
    await async_session.commit()


async def test_bridger(bridger: Bridger, bridge_wrapper: BridgeWrapper, mock_endpoint):
    await bridge_wrapper.enqueue_job(
        bridge=bridger.bridge_name,
        endpoint_name=mock_endpoint,
        job=InferenceJob(
            inference_id="mock",
            payload={
                "data": "mock",
            },
        ),
    )

    await bridger.bridge_all()

    assert await bridger.job_producer.count_jobs(mock_endpoint) == 1



from typing import Any, Awaitable, Callable

from fastapi import Depends

from moriarty.log import logger
from moriarty.matrix.job_manager.bridge.manager import BridgeManager
from moriarty.matrix.job_manager.bridge.plugin import QueueBridge
from moriarty.matrix.job_manager.params import InferenceJob, InferenceResult
from moriarty.tools import ensure_awaitable, sample_as_weights


def get_bridge_manager():
    return BridgeManager()


def get_bridge_wrapper(
    bridge_manager: BridgeManager = Depends(get_bridge_manager),
):
    return BridgeWrapper(bridge_manager)


class BridgeWrapper:
    def __init__(
        self,
        bridge_manager: BridgeManager,
        bridge_kwargs: dict[str, Any] = None,
    ):
        self.bridge_manager = bridge_manager
        self.bridge_kwargs = bridge_kwargs or {}

    def get_bridge(self, bridge_name: str, **kwargs) -> QueueBridge:
        return self.bridge_manager.init(bridge_name, **{**self.bridge_kwargs, **kwargs})

    async def enqueue_job(
        self,
        bridge: str | QueueBridge,
        endpoint_name: str,
        job: InferenceJob,
        priority: int = None,
        *bridge_args,
        **bridge_kwargs,
    ) -> str:
        if isinstance(bridge, str):
            bridge = self.get_bridge(
                bridge,
                *bridge_args,
                **bridge_kwargs,
            )
        return await bridge.enqueue_job(
            endpoint_name=endpoint_name,
            job=job,
            priority=priority,
        )

    async def dequeue_job(
        self,
        bridge: str | QueueBridge,
        endpoint_name: str,
        process_func: Callable[[InferenceJob], None] | Awaitable[InferenceJob],
        size: int = 1,
        priority: int = None,
        *bridge_args,
        **bridge_kwargs,
    ) -> int:
        if isinstance(bridge, str):
            bridge = self.get_bridge(
                bridge,
                *bridge_args,
                **bridge_kwargs,
            )
        process_func = ensure_awaitable(process_func)

        return await bridge.dequeue_job(
            endpoint_name=endpoint_name,
            process_func=process_func,
            size=size,
            priority=priority,
        )

    async def sample_job(
        self,
        bridge: str | QueueBridge,
        endpoint_name: str,
        process_func: Callable[[InferenceJob], None] | Awaitable[InferenceJob],
        *bridge_args,
        **bridge_kwargs,
    ) -> int:
        if isinstance(bridge, str):
            bridge = self.get_bridge(
                bridge,
                *bridge_args,
                **bridge_kwargs,
            )

        avaliable_priorities = await bridge.list_avaliable_priorities(endpoint_name)
        if not avaliable_priorities:
            return 0

        sampled_priority = sample_as_weights(avaliable_priorities)
        logger.info(
            f"Sampled priority: {sampled_priority} from {avaliable_priorities} for {endpoint_name}"
        )
        return await self.dequeue_job(
            bridge,
            endpoint_name=endpoint_name,
            process_func=process_func,
            size=1,
            priority=sampled_priority,
        )

    async def enqueue_result(
        self,
        bridge: str | QueueBridge,
        bridge_result_queue_url: str,
        result: InferenceResult,
        *bridge_args,
        **bridge_kwargs,
    ) -> str:
        if isinstance(bridge, str):
            bridge = self.get_bridge(
                bridge,
                bridge_result_queue_url=bridge_result_queue_url,
                *bridge_args,
                **bridge_kwargs,
            )
        return await bridge.enqueue_result(
            result,
        )

    async def dequeue_result(
        self,
        bridge: str | QueueBridge,
        bridge_result_queue_url: str,
        process_func: Callable[[InferenceResult], None] | Awaitable[InferenceResult],
        size: int = 10,
        *bridge_args,
        **bridge_kwargs,
    ) -> InferenceResult:
        if isinstance(bridge, str):
            bridge = self.get_bridge(
                bridge,
                bridge_result_queue_url=bridge_result_queue_url,
                *bridge_args,
                **bridge_kwargs,
            )
        process_func = ensure_awaitable(process_func)
        return await bridge.dequeue_result(
            process_func,
            size=size,
        )

    async def prune_endpoint(
        self,
        endpoint_name: str,
        bridge: str | QueueBridge,
        *bridge_args,
        **bridge_kwargs,
    ):
        if isinstance(bridge, str):
            bridge = self.get_bridge(
                bridge,
                *bridge_args,
                **bridge_kwargs,
            )
        f = ensure_awaitable(bridge.remove_job_queue)
        await f(endpoint_name)



from urllib.parse import urljoin

import httpx

from moriarty.log import logger
from moriarty.matrix.operator_.enums_ import MetricType
from moriarty.matrix.operator_.params import (
    ContainerScope,
    CreateEndpointParams,
    ListEndpointsResponse,
    QueryEndpointAutoscaleLogResponse,
    QueryEndpointAutoscaleResponse,
    QueryEndpointResponse,
    ResourceScope,
    ScheduleScope,
    SetAutoscaleParams,
    SidecarScope,
    UpdateEndpointParams,
)


def _make_headers(token: str | None) -> dict:
    return {"Authorization": f"Bearer {token}"} if token else dict()


def request_scan_endpoint(
    api_url: str,
    limit: int = None,
    cursor: str = None,
    keyword: str = None,
    order_by: str = None,
    order: str = None,
    token: str = None,
) -> ListEndpointsResponse:
    response = httpx.get(
        urljoin(api_url, "/endpoint/list"),
        params={
            k: v
            for k, v in {
                "limit": limit,
                "cursor": cursor,
                "keyword": keyword,
                "order_by": order_by,
                "order": order,
            }.items()
            if v is not None
        },
        headers=_make_headers(token),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        logger.error(f"Request error, response text: {response.text}")
        raise
    return ListEndpointsResponse.model_validate(response.json())


def request_scan_autoscale_log(
    api_url: str,
    endpoint_name: str,
    limit: int = None,
    cursor: int = None,
    keyword: str = None,
    order_by: str = None,
    order: str = None,
    token: str = None,
) -> QueryEndpointAutoscaleLogResponse:
    response = httpx.get(
        urljoin(api_url, f"/autoscale/{endpoint_name}/log"),
        params={
            k: v
            for k, v in {
                "limit": limit,
                "cursor": cursor,
                "keyword": keyword,
                "order_by": order_by,
                "order": order,
            }.items()
            if v is not None
        },
        headers=_make_headers(token),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        logger.error(f"Request error, response text: {response.text}")
        raise
    return QueryEndpointAutoscaleLogResponse.model_validate(response.json())


def request_query_endpoint(
    endpoint_name: str,
    api_url: str,
    token: str = None,
) -> QueryEndpointResponse:
    response = httpx.get(
        urljoin(api_url, f"/endpoint/{endpoint_name}/info"),
        headers=_make_headers(token),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        logger.error(f"Request error, response: {response.text}")
        raise
    return QueryEndpointResponse.model_validate(response.json())


def request_delete_endpoint(
    endpoint_name: str,
    api_url: str,
    token: str = None,
) -> None:
    response = httpx.post(
        urljoin(api_url, f"/endpoint/{endpoint_name}/delete"),
        headers=_make_headers(token),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        if response.status_code == 404:
            return None

        logger.error(f"Request error, response: {response.text}")
        raise

    return None


def request_delete_autoscale(
    endpoint_name: str,
    api_url: str,
    token: str = None,
) -> None:
    response = httpx.post(
        urljoin(api_url, f"/autoscale/{endpoint_name}/delete"),
        headers=_make_headers(token),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        if response.status_code == 404:
            return None

        logger.error(f"Request error, response: {response.text}")
        raise

    return None


def request_query_autoscale(
    endpoint_name: str,
    api_url: str,
    token: str = None,
) -> QueryEndpointAutoscaleResponse | None:
    response = httpx.get(
        urljoin(api_url, f"/endpoint/{endpoint_name}/autoscale"),
        headers=_make_headers(token),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            return None

        logger.error(f"Request error, response: {response.text}")
        raise
    return QueryEndpointAutoscaleResponse.model_validate(response.json())


def request_set_autoscale(
    endpoint_name: str,
    api_url: str,
    min_replicas: int,
    max_replicas: int,
    scale_in_cooldown: int,
    scale_out_cooldown: int,
    metrics: MetricType,
    metrics_threshold: float,
    token: str = None,
) -> None:
    response = httpx.post(
        urljoin(api_url, f"/autoscale/{endpoint_name}/set"),
        headers=_make_headers(token),
        data=SetAutoscaleParams(
            min_replicas=min_replicas,
            max_replicas=max_replicas,
            scale_in_cooldown=scale_in_cooldown,
            scale_out_cooldown=scale_out_cooldown,
            metrics=metrics,
            metrics_threshold=metrics_threshold,
        ).model_dump_json(),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        logger.error(f"Request error, response: {response.text}")
        raise
    return QueryEndpointResponse.model_validate(response.json())


def request_create_endpoint_with_params(
    api_url,
    params: CreateEndpointParams,
    token: str = None,
) -> None:
    response = httpx.post(
        urljoin(api_url, "/endpoint/create"),
        headers=_make_headers(token),
        data=params.model_dump_json(),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        logger.error(f"Request error, response: {response.text}")
        raise
    return None


def request_update_endpoint_with_params(
    api_url,
    endpoint_name: str,
    params: UpdateEndpointParams,
    token: str = None,
) -> QueryEndpointResponse:
    response = httpx.post(
        urljoin(api_url, f"/endpoint/{endpoint_name}/update"),
        headers=_make_headers(token),
        data=params.model_dump_json(),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        logger.error(f"Request error, response: {response.text}")
        raise
    return QueryEndpointResponse.model_validate(response.json())


def request_create_endpoint(
    api_url: str,
    endpoint_name: str,
    image: str,
    model_path: str = None,
    queue_capacity: int = None,
    replicas: int = None,
    # ResourceScope
    cpu_request: float = None,
    cpu_limit: float = None,
    memory_request: float = None,
    memory_limit: float = None,
    gpu_nums: int = None,
    gpu_type: str = None,
    # ScheduleScope
    node_labels: dict[str, str] = None,
    node_selector: dict[str, str] = None,
    # ContainerScope
    environment_variables: dict[str, str] = None,
    environment_variables_secret_refs: list[str] = None,
    commands: list[str] = None,
    args: list[str] = None,
    invoke_port: int = None,
    invoke_path: str = None,
    health_check_path: str = None,
    # SidecarScope
    concurrency: int = None,
    process_timeout: int = None,
    health_check_timeout: int = None,
    health_check_interval: int = None,
    token: str = None,
) -> None:
    response = httpx.post(
        urljoin(api_url, f"/endpoint/create"),
        headers=_make_headers(token),
        data=CreateEndpointParams(
            **{
                k: v
                for k, v in dict(
                    endpoint_name=endpoint_name,
                    image=image,
                    model_path=model_path,
                    queue_capacity=queue_capacity,
                    replicas=replicas,
                ).items()
                if v is not None
            },
            resource=ResourceScope(
                **{
                    k: v
                    for k, v in dict(
                        cpu_request=cpu_request,
                        cpu_limit=cpu_limit,
                        memory_request=memory_request,
                        memory_limit=memory_limit,
                        gpu_nums=gpu_nums,
                        gpu_type=gpu_type,
                    ).items()
                    if v is not None
                }
            ),
            schedule=ScheduleScope(
                **{
                    k: v
                    for k, v in dict(
                        node_labels=node_labels,
                        node_selector=node_selector,
                    ).items()
                    if v is not None
                }
            ),
            container=ContainerScope(
                **{
                    k: v
                    for k, v in dict(
                        environment_variables=environment_variables,
                        environment_variables_secret_refs=environment_variables_secret_refs,
                        commands=commands,
                        args=args,
                        invoke_port=invoke_port,
                        invoke_path=invoke_path,
                        health_check_path=health_check_path,
                    ).items()
                    if v is not None
                }
            ),
            sidecar=SidecarScope(
                **{
                    k: v
                    for k, v in dict(
                        concurrency=concurrency,
                        process_timeout=process_timeout,
                        health_check_timeout=health_check_timeout,
                        health_check_interval=health_check_interval,
                    ).items()
                    if v is not None
                }
            ),
        ).model_dump_json(),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        logger.error(f"Request error, response: {response.text}")
        raise

    return


def request_update_endpoint(
    api_url: str,
    endpoint_name: str,
    image: str = None,
    model_path: str = None,
    queue_capacity: int = None,
    replicas: int = None,
    need_restart: bool = None,
    # ResourceScope
    cpu_request: float = None,
    cpu_limit: float = None,
    memory_request: float = None,
    memory_limit: float = None,
    gpu_nums: int = None,
    gpu_type: str = None,
    # ScheduleScope
    node_labels: dict[str, str] = None,
    node_selector: dict[str, str] = None,
    # ContainerScope
    environment_variables: dict[str, str] = None,
    environment_variables_secret_refs: list[str] = None,
    commands: list[str] = None,
    args: list[str] = None,
    invoke_port: int = None,
    invoke_path: str = None,
    health_check_path: str = None,
    # SidecarScope
    concurrency: int = None,
    process_timeout: int = None,
    health_check_timeout: int = None,
    health_check_interval: int = None,
    token: str = None,
) -> QueryEndpointResponse:

    response = httpx.post(
        urljoin(api_url, f"/endpoint/{endpoint_name}/update"),
        headers=_make_headers(token),
        data=UpdateEndpointParams(
            **{
                k: v
                for k, v in dict(
                    image=image,
                    model_path=model_path,
                    queue_capacity=queue_capacity,
                    replicas=replicas,
                    need_restart=need_restart,
                ).items()
                if v is not None
            },
            resource=ResourceScope(
                **{
                    k: v
                    for k, v in dict(
                        cpu_request=cpu_request,
                        cpu_limit=cpu_limit,
                        memory_request=memory_request,
                        memory_limit=memory_limit,
                        gpu_nums=gpu_nums,
                        gpu_type=gpu_type,
                    ).items()
                    if v is not None
                }
            ),
            schedule=ScheduleScope(
                **{
                    k: v
                    for k, v in dict(
                        node_labels=node_labels,
                        node_selector=node_selector,
                    ).items()
                    if v is not None
                }
            ),
            container=ContainerScope(
                **{
                    k: v
                    for k, v in dict(
                        environment_variables=environment_variables,
                        environment_variables_secret_refs=environment_variables_secret_refs,
                        commands=commands,
                        args=args,
                        invoke_port=invoke_port,
                        invoke_path=invoke_path,
                        health_check_path=health_check_path,
                    ).items()
                    if v is not None
                }
            ),
            sidecar=SidecarScope(
                **{
                    k: v
                    for k, v in dict(
                        concurrency=concurrency,
                        process_timeout=process_timeout,
                        health_check_timeout=health_check_timeout,
                        health_check_interval=health_check_interval,
                    ).items()
                    if v is not None
                }
            ),
        ).model_dump_json(),
        follow_redirects=True,
    )
    try:
        response.raise_for_status()
    except httpx.HTTPStatusError:
        logger.error(f"Request error, response: {response.text}")
        raise

    return QueryEndpointResponse.model_validate(response.json())



import asyncio
import os

from moriarty.sidecar.producer import JobProducer


async def send_stop_sig(pid, wait=2):
    await asyncio.sleep(wait)
    os.kill(pid, 15)


async def test_sidecar(async_redis_client, endpoint_name, inference_consumer):
    producer = JobProducer(redis_client=async_redis_client)

    invoke_params = {"inference_id": "hello world"}
    pid = os.getpid()
    asyncio.create_task(send_stop_sig(pid))
    await producer.invoke(endpoint_name, invoke_params)
    await inference_consumer.run_forever()
    assert inference_consumer._last_response == invoke_params



import os
from contextlib import asynccontextmanager

from fastapi import Depends, FastAPI, HTTPException, Query, Response

from moriarty.log import logger
from moriarty.matrix.operator_.operator_ import Operator, get_operaotr
from moriarty.sidecar.params import MatrixCallback


@asynccontextmanager
async def lifespan(app: FastAPI):
    yield


app = FastAPI(lifespan=lifespan)


@app.post("/callback")
async def callback(
    callback: MatrixCallback,
    operator: Operator = Depends(get_operaotr),
) -> Response:
    await operator.handle_callback(callback)
    return Response(status_code=200)


@app.get("/")
async def root():
    return {"message": "Hello World"}



import asyncio
import sys
from functools import partial
from urllib.parse import urljoin

import httpx
import redis.asyncio as redis
from brq.consumer import Consumer
from brq.daemon import Daemon

from moriarty.log import logger
from moriarty.sidecar.params import InferenceProxyStatus, MatrixCallback

if sys.version_info >= (3, 11):
    import asyncio as async_timeout
else:
    import async_timeout

from .vars import GROUP_NAME, REDIS_PREFIX


class InferencerConsumer:
    REDIS_PREFIX = REDIS_PREFIX
    GROUP_NAME = GROUP_NAME

    def __init__(
        self,
        redis_client: redis.Redis | redis.RedisCluster,
        endpoint_name: str,
        invoke_host: str = "localhost",
        invoke_port: int = "8080",
        invoke_path: str = "/invocations",
        health_check_path: str = "/ping",
        callbacl_url: str = "http://moriarty-operator-callback:8999/callback",
        enable_retry: bool = False,
        enable_ssl: bool = False,
        concurrency: int = 1,
        process_timeout: int = 3600,
        health_check_timeout: int = 1200,
        health_check_interval: int = 5,
        **kwargs,
    ):
        self.redis_client = redis_client
        self.process_timeout = process_timeout
        self.health_check_timeout = health_check_timeout
        self.health_check_interval = health_check_interval

        scheme = "https" if enable_ssl else "http"
        proxy_url = f"{scheme}://{invoke_host}:{invoke_port}"
        self.invoke_url = urljoin(proxy_url, invoke_path)
        self.health_check_path = urljoin(proxy_url, health_check_path)
        self.callback_url = callbacl_url

        self._consumer_builder = partial(
            Consumer,
            self.redis_client,
            self._proxy,
            register_function_name=endpoint_name,
            redis_prefix=self.REDIS_PREFIX,
            group_name=self.GROUP_NAME,
            enable_enque_deferred_job=False,  # No deferred job for inferencer
            enable_reprocess_timeout_job=enable_retry,
            enable_dead_queue=False,
            process_timeout=self.process_timeout,
            delete_message_after_process=True,
            **kwargs,
        )
        self.daemon = Daemon(*[self._consumer_builder() for _ in range(concurrency)])

    async def _proxy(self, payload: dict) -> dict | None:
        inference_id = payload.get("inference_id") or payload.get("inferenceId")
        if not inference_id:
            return await self.return_no_inference_id_error(payload)
        async with httpx.AsyncClient() as client:
            logger.debug(f"Invoke endpoint: {self.invoke_url}")
            try:
                response = await client.post(
                    self.invoke_url,
                    json=payload,
                    timeout=self.process_timeout,
                    follow_redirects=True,
                )
            except httpx.ConnectError as e:
                logger.error(f"Invoke endpoint failed: {e}")
                logger.exception(e)
                await self._callback(MatrixCallback.from_exception(inference_id, e))
                return None
            else:
                logger.debug(f"Invoke endpoint response status: {response.status_code}")
                await self._callback(MatrixCallback.from_response(inference_id, response))
                return response.json()

    async def return_no_inference_id_error(self, payload: dict) -> None:
        await self._callback(
            MatrixCallback(
                status=InferenceProxyStatus.INTERNAL_ERROR,
                msg=f"Missing 'inference_id' or 'inferenceId' in payload: {payload}",
            )
        )

    async def _callback(self, callback_params: MatrixCallback) -> None:
        async with httpx.AsyncClient(follow_redirects=True) as client:
            try:
                await client.post(
                    self.callback_url,
                    data=callback_params.model_dump_json(),
                    timeout=60,
                )
            except httpx.ConnectError as e:
                logger.error(f"Callback failed: {e}")

    async def run_forever(self) -> None:
        try:
            async with async_timeout.timeout(self.health_check_timeout):
                logger.info(
                    f"Wait for invoke endpoint online...timeout: {self.health_check_timeout}s"
                )
                await self.wait_for_health_check()
        except TimeoutError:
            logger.error(f"Invoke endpoint not online in {self.health_check_timeout}s. Exit...")
            raise

        logger.info(f"Healthy check passed. Start consumer...")
        await self.daemon.run_forever()

    async def wait_for_health_check(self) -> None:
        async with httpx.AsyncClient() as client:
            while True:
                try:
                    await client.get(self.health_check_path, timeout=60)
                    logger.debug(f"Ping {self.health_check_path} succeeded")
                    break
                except httpx.ConnectError as e:
                    logger.debug(f"Ping {self.health_check_path} failed: {e}")
                    await asyncio.sleep(self.health_check_interval)



import socket
import subprocess
import time
from pathlib import Path
from typing import Coroutine

import httpx
import pytest

import docker
from moriarty.sidecar.consumer import InferencerConsumer
from moriarty.sidecar.params import MatrixCallback
from moriarty.sidecar.producer import JobProducer

_HERE = Path(__file__).parent


def get_port():
    # Get an unoccupied port
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


@pytest.fixture(scope="session")
def docker_client():
    try:
        client = docker.from_env()
        client.ping()
        return client
    except:
        pytest.skip("Docker is not available")


@pytest.fixture(scope="session")
def inference_port():
    inference_port = get_port()
    try:
        p = subprocess.Popen(
            [
                "uvicorn",
                "app:app",
                "--host",
                "0.0.0.0",
                "--port",
                str(inference_port),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=_HERE / "mock",
        )
        time.sleep(1)  # Wait for the server to start
        for _ in range(10):
            try:
                if p.poll() is not None:
                    raise TimeoutError(
                        f"Inference server not ready in 5s. stdout: {p.stdout} stderr: {p.stderr}"
                    )

                httpx.get(f"http://127.0.0.1:{inference_port}/ping")
            except:
                time.sleep(0.5)
            else:
                break
        else:
            p.kill()
            raise TimeoutError(
                f"Inference server not ready in 5s. stdout: {p.stdout} stderr: {p.stderr}"
            )
        yield inference_port
    finally:
        p.kill()


@pytest.fixture
def endpoint_name(case_id):
    return f"unittest-{case_id}"


class MockInferenceConsumer(InferencerConsumer):
    async def _callback(self, *args, **kwargs):
        # Disable callback
        return

    async def _proxy(self, *args, **kwargs):
        self._last_response = await super()._proxy(*args, **kwargs)
        return self._last_response


@pytest.fixture
async def inference_consumer(async_redis_client, endpoint_name, inference_port):
    return MockInferenceConsumer(
        redis_client=async_redis_client,
        endpoint_name=endpoint_name,
        invoke_port=inference_port,
    )



import asyncio
import contextlib
import signal

import redis.asyncio as redis
from sqlalchemy.ext.asyncio import AsyncSession

from moriarty.envs import get_bridge_name
from moriarty.log import logger
from moriarty.matrix.job_manager.bridge_wrapper import (
    get_bridge_manager,
    get_bridge_wrapper,
)
from moriarty.matrix.operator_.autoscaler import AutoscalerManager
from moriarty.matrix.operator_.config import Config
from moriarty.matrix.operator_.dbutils import open_db_session
from moriarty.matrix.operator_.operator_ import Bridger
from moriarty.matrix.operator_.rds import open_redis_client
from moriarty.matrix.operator_.spawner.impl.k8s import KubeSpawner
from moriarty.matrix.operator_.spawner.manager import (
    SpawnerManager,
    get_spawner,
    get_spawner_manager,
    get_spawner_name,
)
from moriarty.matrix.operator_.spawner.plugin import Spawner
from moriarty.matrix.operator_.tools import load_kube_config


async def event_wait(evt, timeout):
    # suppress TimeoutError because we'll return False in case of timeout
    with contextlib.suppress(asyncio.TimeoutError):
        await asyncio.wait_for(evt.wait(), timeout)
    return evt.is_set()


class DaemonMixin:
    interval = 0.1

    def __init__(self) -> None:
        self._stop_event = asyncio.Event()
        self._task: None | asyncio.Task = None

    async def initialize(self):
        pass

    async def run(self):
        pass

    async def cleanup(self):
        pass

    async def _start(self):
        if self._task is not None:
            raise RuntimeError("Already running")

        self._stop_event.clear()
        self._task = asyncio.create_task(self._task_wrapper())

    async def _stop(self):
        self._stop_event.set()
        if self._task is not None:
            logger.info("Waiting for task to finish...")
            await self._task

    async def _task_wrapper(self):
        try:
            await self.initialize()
        except Exception as e:
            # Unexpected error, stop the autoscaler
            logger.exception(e)
            exit(1)

        await self.run()
        while not await event_wait(self._stop_event, self.interval):
            if self._stop_event.is_set():
                break
            try:
                await self.run()
            except Exception as e:
                logger.exception(e)

        await self.cleanup()

    async def run_forever(self, stop_signals: list = [signal.SIGINT, signal.SIGTERM]):
        loop = asyncio.get_event_loop()

        stop_event = asyncio.Event()

        async def _stop():
            logger.debug("Signal received")
            stop_event.set()

        for sig in stop_signals:
            loop.add_signal_handler(sig, lambda: asyncio.create_task(_stop()))

        logger.info(f"Starting daemon...")
        await self._start()
        logger.info(f"Daemon started, waiting for signals {stop_signals}...")
        await stop_event.wait()

        logger.info(f"Terminating daemon...")
        await self._stop()
        logger.info(f"Daemon terminated")


class KubeAutoscalerDaemon(DaemonMixin):
    interval = 60

    def __init__(self, config: Config) -> None:
        super().__init__()

        self.spawner: KubeSpawner = None
        self.spawner_manager = SpawnerManager()
        self.config = config

    async def initialize(self):
        self.spawner = await get_spawner(
            spawner_manager=get_spawner_manager(),
            spawner_name=get_spawner_name(),
        )

    async def run(self):
        logger.info(f"Triggering autoscaler...")
        async with open_redis_client(self.config) as redis_client:
            async with open_db_session(self.config) as session:
                autoscaler = AutoscalerManager(
                    redis_client=redis_client, session=session, spawner=self.spawner
                )
                await autoscaler.scan_and_scale()


class BridgeDaemon(DaemonMixin):
    interval = 3

    def __init__(self, config: Config) -> None:
        super().__init__()
        self.bridge_name = get_bridge_name()
        self.bridge_wrapper = get_bridge_wrapper(bridge_manager=get_bridge_manager())
        self.config = config
        self.spawner: Spawner = None

    async def initialize(self):
        self.spawner = await get_spawner(
            spawner_manager=get_spawner_manager(),
            spawner_name=get_spawner_name(),
        )

    async def run(self):
        logger.debug(f"Triggering bridge...")

        async with open_redis_client(self.config) as redis_client:
            async with open_db_session(self.config) as session:
                bridger = Bridger(
                    spawner=self.spawner,
                    bridge_name=self.bridge_name,
                    bridge_wrapper=self.bridge_wrapper,
                    redis_client=redis_client,
                    session=session,
                )
                await bridger.bridge_all()
